#include <stdlib.h>
#include <string.h>
#include <pspkernel.h>
#include "display.h"

static byte * vram_base = NULL;
byte * vram_start = NULL;
static bool vram_page = 0;
static byte * cfont_buffer = NULL, * efont_buffer = NULL;

#define DISP_EFONTSIZE (DISP_FONTSIZE)
#define DISP_CFONTSIZE (DISP_FONTSIZE * DISP_FONTSIZE / 8)
#define DISP_RSPAN 0

extern void disp_init()
{
	sceDisplaySetMode(0, 480, 272);
	vram_page = 0;
	vram_base = (byte *)0x04000000;
	vram_start = (byte *)0x44044000;
	sceDisplaySetFrameBuf(vram_base, 512, PSP_DISPLAY_PIXEL_FORMAT_5551, PSP_DISPLAY_SETBUF_NEXTFRAME);
}

extern bool disp_load_font(const char * efont, const char * cfont)
{
	int size;
	int fd = sceIoOpen(efont, PSP_O_RDONLY, 0777);
	if(fd < 0)
		return false;
	size = sceIoLseek32(fd, 0, PSP_SEEK_END);
	if((efont_buffer = (byte *)calloc(1, size)) == NULL)
	{
		sceIoClose(fd);
		return false;
	}
	sceIoLseek32(fd, 0, PSP_SEEK_SET);
	sceIoRead(fd, efont_buffer, size);
	sceIoClose(fd);

	fd = sceIoOpen(cfont, PSP_O_RDONLY, 0777);
	if(fd < 0)
	{
		disp_free_font();
		return false;
	}
	size = sceIoLseek32(fd, 0, PSP_SEEK_END);
	if((cfont_buffer = (byte *)calloc(1, size)) == NULL)
	{
		disp_free_font();
		sceIoClose(fd);
		return false;
	}
	sceIoLseek32(fd, 0, PSP_SEEK_SET);
	sceIoRead(fd, cfont_buffer, size);
	sceIoClose(fd);
	return true;
}

extern void disp_free_font()
{
	if(efont_buffer != NULL)
		free((void *)efont_buffer);
	if(cfont_buffer != NULL)
		free((void *)cfont_buffer);
}

extern void disp_flip()
{
	disp_waitv();
	vram_page ^= 1;
	vram_base = (byte *)0x04000000 + (vram_page ? 0x44000 : 0);
	vram_start = (byte *)0x44000000 + (vram_page ? 0 : 0x44000);
	sceDisplaySetFrameBuf(vram_base, 512, PSP_DISPLAY_PIXEL_FORMAT_5551, PSP_DISPLAY_SETBUF_IMMEDIATE);
}

extern void disp_getimage(dword x, dword y, dword w, dword h, word * buf)
{
	byte * lines = vram_base + 0x40000000, * linesend = lines + (min(272 - y, h) << 10);
	dword rw = (min(512 - x, w) << 1);
	for(;lines < linesend; lines += 1024)
	{
		memcpy(buf, lines, rw);
		buf += w;
	}
}

extern void disp_putimage(dword x, dword y, dword w, dword h, dword startx, dword starty, word * buf)
{
	byte * lines = disp_get_vaddr(x, y), * linesend = lines + (min(272 - y, h - starty) << 10);
	buf = buf + starty * w + startx;
	dword rw = (min(512 - x, w - startx) << 1);
	for(;lines < linesend; lines += 1024)
	{
		memcpy(lines, buf, rw);
		buf += w;
	}
}

extern void disp_duptocache()
{
	memmove(vram_start, vram_base + 0x40000000, 0x44000);
}

extern void disp_rectduptocache(dword x1, dword y1, dword x2, dword y2)
{
	byte * lines = disp_get_vaddr(x1, y1), * linesend = disp_get_vaddr(x1, y2), * lined = vram_base + 0x40000000 + (x1 << 1) + (y1 << 10);
	dword w = (x2 - x1 + 1) * 2;
	for(;lines <= linesend; lines += 1024, lined += 1024)
		memcpy(lines, lined, w);
}

extern void disp_putnstring(word x, word y, word color, const byte *str, int count)
{
	byte * vaddr;
	const byte * ccur, * cend;

	while(*str != 0 && count > 0)
	{
		if(*str > 0x80)
		{
			if(x > 480 - DISP_RSPAN - DISP_FONTSIZE)
			{
				x = 0;
				y += DISP_FONTSIZE;
			}
			vaddr = disp_get_vaddr(x,y);
			//Hyunee add start
			if (((*str>=0x80) && (*str<0xA0)) || (*str>=0xA1)) {
				ccur=cfont_buffer+((dword)(*str - 0xA1) * 0x5e + (dword)(*(str + 1) -0xA1)) * DISP_CFONTSIZE;
			}else{ // original code
				ccur = cfont_buffer + (((dword)(*str - 0x81)) * 0xBF + ((dword)(*(str + 1) - 0x40))) * DISP_CFONTSIZE;
			}
			//Hyunee add end
		
			for (cend = ccur + DISP_CFONTSIZE; ccur < cend; ccur ++) {
				int b;
				byte ch =*ccur;
				byte * vpoint = vaddr;
				for (b = 0x80; b > 0; b >>= 1) {
					if ((ch & b) != 0)
						*(word *)vpoint = color;
					vpoint += 2;
				}
#ifdef FONT12
				ch = *(++ ccur);
				for (b = 0x80; b > 8; b >>= 1) {
					if ((ch & b) != 0)
						*(word *)vpoint = color;
					vpoint += 2;
				}
				vaddr += 1024;
				vpoint = vaddr;
				for (; b > 0; b >>= 1) {
					if ((ch & b) != 0)
						*(word *)vpoint = color;
					vpoint += 2;
				}
#endif
				ch = *(++ ccur);
				for (b = 0x80; b > 0; b >>= 1) {
					if ((ch & b) != 0)
						*(word *)vpoint = color;
					vpoint += 2;
				}
				vaddr += 1024;
			}
			str += 2;
			count -= 2;
			x += DISP_FONTSIZE;
		}
		else if(*str > 0x1F)
		{
			if(x > 480 - DISP_RSPAN - DISP_FONTSIZE / 2)
			{
				x = 0;
				y += DISP_FONTSIZE;
			}
			vaddr = disp_get_vaddr(x,y);
#ifdef FONT12
			ccur = efont_buffer + (((dword)*str) - 0x20) * DISP_EFONTSIZE;
#else
			ccur = efont_buffer + ((dword)*str) * DISP_EFONTSIZE;
#endif

			for (cend = ccur + DISP_EFONTSIZE; ccur < cend; ccur ++) {
				int b;
				byte * vpoint = vaddr;
#ifdef FONT12
				for (b = 0x80; b > 2; b >>= 1) {
#else
				for (b = 0x80; b > 0; b >>= 1) {
#endif
					if (((*ccur) & b) != 0)
						*(word *)vpoint = color;
					vpoint += 2;
				}
				vaddr += 1024;
			}
			str ++;
			count --;
			x += DISP_FONTSIZE / 2;
		}
		else
		{
			if(x > 480 - DISP_RSPAN - DISP_FONTSIZE / 2)
			{
				x = 0;
				y += DISP_FONTSIZE;
			}
			str ++;
			count --;
			x += DISP_FONTSIZE / 2;
		}
	}
}

extern void disp_putnstringvert(word x, word y, word color, const byte *str, int count)
{
	byte * vaddr;
	const byte * ccur, * cend;

	while(*str != 0 && count > 0)
	{
		if(*str > 0x80)
		{
			if(y > 272 - DISP_RSPAN - DISP_FONTSIZE)
			{
				y = 0;
				x -= DISP_FONTSIZE;
			}
			vaddr = disp_get_vaddr(x,y);

			//Hyunee add start
			if (((*str>=0x80) && (*str<0xA0)) || (*str>=0xA1)) {
				ccur=cfont_buffer+((dword)(*str - 0xA1) * 0x5e + (dword)(*(str + 1) -0xA1)) * DISP_CFONTSIZE;
			}else{ // original code
				ccur = cfont_buffer + (((dword)(*str - 0x81)) * 0xBF + ((dword)(*(str + 1) - 0x40))) * DISP_CFONTSIZE;
			}
			//Hyunee add end
		
			for (cend = ccur + DISP_CFONTSIZE; ccur < cend; ccur ++) {
				int b;
				byte ch =*ccur;
				byte * vpoint = vaddr;
				for (b = 0x80; b > 0; b >>= 1) {
					if ((ch & b) != 0)
						*(word *)vpoint = color;
					vpoint += 1024;
				}
#ifdef FONT12
				ch = *(++ ccur);
				for (b = 0x80; b > 8; b >>= 1) {
					if ((ch & b) != 0)
						*(word *)vpoint = color;
					vpoint += 1024;
				}
				vaddr -= 2;
				vpoint = vaddr;
				for (; b > 0; b >>= 1) {
					if ((ch & b) != 0)
						*(word *)vpoint = color;
					vpoint += 1024;
				}
#endif
				ch = *(++ ccur);
				for (b = 0x80; b > 0; b >>= 1) {
					if ((ch & b) != 0)
						*(word *)vpoint = color;
					vpoint += 1024;
				}
				vaddr -= 2;
			}
			str += 2;
			count -= 2;
			y += DISP_FONTSIZE;
		}
		else if(*str > 0x1F)
		{
			if(y > 272 - DISP_RSPAN - DISP_FONTSIZE / 2)
			{
				y = 0;
				x -= DISP_FONTSIZE;
			}
			vaddr = disp_get_vaddr(x,y);
#ifdef FONT12
			ccur = efont_buffer + (((dword)*str) - 0x20) * DISP_EFONTSIZE;
#else
			ccur = efont_buffer + ((dword)*str) * DISP_EFONTSIZE;
#endif

			for (cend = ccur + DISP_EFONTSIZE; ccur < cend; ccur ++) {
				int b;
				byte * vpoint = vaddr;
#ifdef FONT12
				for (b = 0x80; b > 2; b >>= 1) {
#else
				for (b = 0x80; b > 0; b >>= 1) {
#endif
					if (((*ccur) & b) != 0)
						*(word *)vpoint = color;
					vpoint += 1024;
				}
				vaddr -= 2;
			}
			str ++;
			count --;
			y += DISP_FONTSIZE / 2;
		}
		else
		{
			if(y > 272 - DISP_RSPAN - DISP_FONTSIZE / 2)
			{
				y = 0;
				x -= DISP_FONTSIZE;
			}
			str ++;
			count --;
			y += DISP_FONTSIZE / 2;
		}
	}
}

extern void disp_fillvram(dword color)
{
	byte *vram, *vram_end;		//pointer to vram

	if ((color >> 8) == (color & 0xFF))
	{
		color = (color & 0xFF);
		memset(vram_start, color, 0x44000);
	}
	else
		for (vram = vram_start, vram_end = vram + 0x44000; vram < vram_end; vram += 2)
			*(word *)vram = color;
}

extern void disp_fillrect(dword x1, dword y1, dword x2, dword y2, dword color)
{
	byte * vsram, * vsram_end, * vram, * vram_end;
	dword wbytes;
	if(x1 > x2)
	{
		dword t = x1; x1 = x2; x2 = t;
	}
	if(y1 > y2)
	{
		dword t = y1; y1 = y2; y2 = t;
	}
	vsram = disp_get_vaddr(x1, y1);
	vsram_end = vsram + 1024 * (y2 - y1);
	if ((color >> 8) == (color & 0xFF))
	{
		color = (color & 0xFF);
		wbytes = (x2 - x1 + 1) * 2;
		for(;vsram <= vsram_end; vsram += 1024)
			memset(vsram, color, wbytes);
	}
	else
	{
		wbytes = (x2 - x1) * 2;
		for(;vsram <= vsram_end; vsram += 1024)
			for(vram = vsram, vram_end = vram + wbytes; vram <= vram_end; vram += 2)
				*(word *)vram = color;
	}
}

extern void disp_rectangle(dword x1, dword y1, dword x2, dword y2, dword color)
{
	byte *vsram, * vram, * vram_end;
	if(x1 > x2)
	{
		dword t = x1; x1 = x2; x2 = t;
	}
	if(y1 > y2)
	{
		dword t = y1; y1 = y2; y2 = t;
	}
	vsram = disp_get_vaddr(x1, y1);
	for(vram = vsram, vram_end = vram + 2 * (x2 - x1); vram < vram_end; vram += 2)
		*(word *)vram = color;
	for(vram_end = vram + 1024 * (y2 - y1); vram < vram_end; vram += 1024)
		*(word *)vram = color;
	for(vram = vsram, vram_end = vram + 1024 * (y2 - y1); vram < vram_end; vram += 1024)
		*(word *)vram = color;
	for(vram_end = vram + 2 * (x2 - x1); vram < vram_end; vram += 2)
		*(word *)vram = color;
	*(word *)vram = color;
}

extern void disp_line(dword x1, dword y1, dword x2, dword y2, dword color)
{
	byte * vram;
	int dy, dx, x, y, d;
	dx = (int)x2 - (int)x1;
	dy = (int)y2 - (int)y1;
	if(dx < 0)
		dx = -dx;
	if(dy < 0)
		dy = -dy;
	d = -dx;
	x = (int)x1;
	y = (int)y1;
	if(dx >= dy)
	{
		if(y2 < y1)
		{
			dword t = x1; x1 = x2; x2 = t; t = y1; y1 = y2; y2 = t;
		}
		vram = disp_get_vaddr(x1, y1);
		if(x1 < x2)
		{
			for(x = x1; x <= x2; x ++)
			{
				if(d > 0)
				{
					y ++;
					vram += 1024;
					d -= 2 * dx;
				}
				d += 2 * dy;
				*(word *)vram = color;
				vram += 2;
			}
		}
		else
		{
			for(x = x1; x >= x2; x --)
			{
				if(d > 0)
				{
					y ++;
					vram += 1024;
					d -= 2 * dx;
				}
				d += 2 * dy;
				*(word *)vram = color;
				vram -= 2;
			}
		}
	}
	else
	{
		if(x2 < x1)
		{
			dword t = x1; x1 = x2; x2 = t; t = y1; y1 = y2; y2 = t;
		}
		vram = disp_get_vaddr(x1, y1);
		if(y1 < y2)
		{
			for(y = y1; y <= y2; y ++)
			{
				if(d > 0)
				{
					x ++;
					vram += 2;
					d -= 2 * dy;
				}
				d += 2 * dx;
				*(word *)vram = color;
				vram += 1024;
			}
		}
		else
		{
			for(y = y1; y >= y2; y --)
			{
				if(d > 0)
				{
					x ++;
					vram += 2;
					d -= 2 * dy;
				}
				d += 2 * dx;
				*(word *)vram = color;
				vram -= 1024;
			}
		}
	}
}
