#ifndef _DISPLAY_H_
#define _DISPLAY_H_

#include <pspdisplay.h>
#include "common/datatype.h"

#ifdef FONT12
#define DISP_FONTSIZE 12
#else
#define DISP_FONTSIZE 16
#endif

extern byte * vram_start;

// R,G,B color to word value color
#define RGB(r,g,b) ((((word)(b) & 0xF8)<<7)+(((word)(g) & 0xF8)<<2)+(((word)(r) & 0xF8)>>3)+0x8000)
#define RGB2(r,g,b) ((((word)(b))<<10)+(((word)(g))<<5)+(((word)(r)))+0x8000)
#define RGB_R(c) ((c) & 0x1F)
#define RGB_G(c) (((c) >> 5) & 0x1F)
#define RGB_B(c) (((c) >> 10) & 0x1F)

// sceDisplayWaitVblankStart function alias name, define is faster than function call (even at most time this is inline linked)
#define disp_waitv() sceDisplayWaitVblankStart()

#define disp_get_vaddr(x, y) (vram_start + ((x) << 1) + ((y) << 10))

#define disp_putpixel(x,y,c) *(word*)disp_get_vaddr((x),(y)) = (c)

extern void disp_init();

extern bool disp_load_font(const char * efont, const char * cfont);

extern void disp_free_font();

extern void disp_flip();

extern void disp_getimage(dword x, dword y, dword w, dword h, word * buf);

extern void disp_putimage(dword x, dword y, dword w, dword h, dword startx, dword starty, word * buf);

extern void disp_duptocache();

extern void disp_rectduptocache(dword x1, dword y1, dword x2, dword y2);

extern void disp_putnstring(word x, word y, word color, const byte *str, int count);
#define disp_putstring(x,y,color,str) disp_putnstring((x),(y),(color),(str),0x7FFFFFFF)

extern void disp_putnstringvert(word x, word y, word color, const byte *str, int count);
#define disp_putstringvert(x,y,color,str) disp_putnstringvert((x),(y),(color),(str),0x7FFFFFFF)

extern void disp_fillvram(dword color);

extern void disp_fillrect(dword x1, dword y1, dword x2, dword y2, dword color);

extern void disp_rectangle(dword x1, dword y1, dword x2, dword y2, dword color);

extern void disp_line(dword x1, dword y1, dword x2, dword y2, dword color);

#endif
