#include <psptypes.h>
#include <pspdebug.h>
#include <string.h>

#include <stdlib.h>
#include <pspkernel.h>
#include <pspdisplay.h>

#include "font_h11.c"
#include "font_e11.c"

#include "HDrawText.h"

#define disp_get_vaddr(x, y) (g_vram_start + ((x) << 1) + ((y) << 10))

int iFontWidth = 11;         // Font Width
int iFontHeight = 12;        // Font Height
int iFToFSpaceWidth = 1;     // Font   Width
int iFToFSpaceHeight = 0;    // Font   Height
int iEngFontWidth = 6 ;      //  Font Width
int iEngFToFSpaceWidth = 0;  //  Font   Width
int iPSPFontWidth = 79;      // PSP ¿ font 
int iPSPFontHeight = 22;     // PSP  font 

//static u32* g_vram_base = (u32 *) 0x04000000;
static u16* g_vram_base = (u16 *) 0x04000000;
unsigned char* g_vram_start  = (unsigned char *) 0x44044000;
static int g_vram_page = 0;
static int g_vram_offset = 0;

void HFont_disp_flip()
{
	g_vram_page ^= 1;
	g_vram_base = (unsigned char *)0x04000000 + (g_vram_page ? 0x44000 : 0);
	g_vram_start = (unsigned char *)0x44000000 + (g_vram_page ? 0 : 0x44000);
}


unsigned short *GetHangulFont(unsigned char u,unsigned char d) {
	// mapping
	u -= 0xB0;
	d -= 0xA1;

	return (unsigned short *)&hangle_font_11[((u*0x5e)+d)*  iFontHeight];
}

unsigned short *GetExtFont(unsigned char u,unsigned char d) {
	// mapping
	u -= 0xA1;
	d -= 0xA1;
	return (unsigned short *)&ext_font_16[((u*0x5e)+d) * 16];
}

unsigned char *GetEnglishFont(unsigned char ch) {
	// mapping
	if      (ch<0x20) ch  = 0;
	else if (ch<0x80) ch -= 0x20;
	else if (ch<0xa0) ch  = 0;
	else              ch -= 0x40;

	return (unsigned char *)&english_font_11[ch*iFontHeight];
}

// ---------------------------------------------------------------------------------

#define PSP_LINE_SIZE 512

void Draw_Char_Ext(int x,int y,unsigned char u,unsigned char d,int color,int bg_col) {
	int i,j;
	unsigned short *fnt;

	u16 *vram;
	u16 *vram_ptr;

//	vram = g_vram_base + x;
//	vram += (y * PSP_LINE_SIZE);
	vram = disp_get_vaddr(x, y);

	fnt = GetExtFont(u, d);

	for (i=0; i<16; i++, fnt++) {
		vram_ptr  = vram;

		for (j=0; j<iFontWidth; j++) {
			if ((*fnt & (65536 >> j)))  *vram_ptr = color;
			*vram_ptr++;
		}

    vram += PSP_LINE_SIZE;
	}
}

void Draw_Char_Hangul(int x,int y,unsigned char u,unsigned char d,int color,int bg_col) {
	
	int i,j;
	unsigned short *fnt;

//	u32 *vram;
//	u32 *vram_ptr;
	u16 *vram;
	u16 *vram_ptr;

//	vram = g_vram_base + x;
//	vram += (y * PSP_LINE_SIZE);
	vram = disp_get_vaddr(x, y);

	fnt = GetHangulFont(u, d);

	for (i=0; i<iFontHeight; i++, fnt++) {
		vram_ptr  = vram;

		for (j=0; j<iFontWidth; j++) {
			if ((*fnt & (65536 >> j)))  *vram_ptr = color;
			*vram_ptr++;
		}

    vram += PSP_LINE_SIZE;
  }
}

void Draw_Char_English(int x,int y,unsigned char ch,int color,int bg_col) {

	int 	i,j;
	unsigned char  *fnt;

//	u32 *vram;
//	u32 *vram_ptr;
	u16 *vram;
	u16 *vram_ptr;

//	vram = g_vram_base + x;
//	vram += (y * PSP_LINE_SIZE);
	vram = disp_get_vaddr(x, y);
	
	fnt = GetEnglishFont(ch);
	
	for (i=0; i<iFontHeight; i++, fnt++) {
		
		vram_ptr  = vram;
		for (j=0; j<iEngFontWidth; j++) {
			if ((*fnt & (128 >> j))) *vram_ptr = color; 
			*vram_ptr++ ;
		}
    vram += PSP_LINE_SIZE;
  }
}

int HDrawText(int x,int y,char *buf,int col,int backcol, int fill) {
	return HDrawTextLimit(x, y, buf, col, backcol, fill, 0, strlen(buf));
}

int HDrawTextLimit(int x,int y,char *buf,int col,int backcol, int fill, int offset, int length) {
	unsigned char ch = 0,bef = 0;
	int pr_index = 0;
	unsigned char *str;

	str = &buf[offset];

	while(*str != 0) {
		ch = *str++;
		pr_index++;
		if (bef!=0) {
			if (bef==0x88 && ch==0x81) {

//			} else {
			} else if(bef==0xA1 || bef==0xA2 || bef==0xA3){
				Draw_Char_Ext(x,y,bef,ch,col,backcol);
				x+=(iFontWidth+iFToFSpaceWidth);
			} else {
				Draw_Char_Hangul(x,y,bef,ch,col,backcol);
				x+=(iFontWidth+iFToFSpaceWidth);
			}
			bef=0;
		} else {
			if (ch==0x0d) {
				ch = *str++;
				if (ch==0x0a)  pr_index++;
				break;
			} else if (ch==0x0a) {
				break;
			} else if (((ch>=0x80) && (ch<0xa0)) || (ch>=0xB0)) {
				bef = ch;
			} else {
				Draw_Char_English(x,y,ch,col,backcol);
				x+=(iEngFontWidth+iEngFToFSpaceWidth);
			}
		}
		if (pr_index>=length) break;
	}

	if (bef!=0)  pr_index--;

	return pr_index;
}

